/*
 GAMP - (c) 2001 - Renato R. Santana <rrsantana@yahoo.com>
 
 This contains MOD playing and controling routines.
   
 This file is part of GAMP, the GNU Amiga Module Player
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
    
 You should have received a copy of the GNU General Public License
 along with This program; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*#define USE_THREAD 1*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <linux/soundcard.h>
#include <sys/resource.h>

#ifdef USE_THREAD
#include <pthread.h>
#include <semaphore.h>
#include <sched.h>
#endif

#include "effects.h"
#include "modctl.h"
#include "mod.h"
#include "driver.h"

#define BUFSIZE						2048
#ifdef USE_THREAD
sem_t								sem_bin;
sem_t								sem_bin2;
	
struct sched_param					scheduling_value;
pthread_t							snd_thread;
pthread_attr_t						mp_thread_attr;
void								*thread_result;

#endif

struct driver       				oss_driver;

//static unsigned int 				blockbuf[BUFSIZE];
//static unsigned char 				buf2[BUFSIZE];
static unsigned int						*blockbuf;
static unsigned char						*buf2;
extern long int						fragsize;
int									mp_playing=0;
struct modheadertype				mod_header;
extern struct modpattern           	patterns[128];
struct channeltype					channel1;
struct channeltype					channel2;
struct channeltype					channel3;
struct channeltype					channel4; 
struct modctl						modctl;
int									bufisfull = 0;
char								blob[5] = " _-=";
int									tempo=6;

extern struct driver				oss_driver;

void process_effect(struct channeltype *channel, int ef_num, int ef_arg, int period);
void refresh_channel_row(	struct channeltype *channel1, 
							struct modheadertype *mod_header, 
							unsigned char modchannel[]);
void *channel_mix(struct channeltype *channel1, int bufpos);
void *snd_play(void *arg);
void *mix_to_buffer(void *arg);
void mp_stop(void);
void mp_setvolume(int vol);

void *channel_mix(struct channeltype *ch, int bufidx){
	register int		i;		/* instrument data index */
	register int		d;		/* instrument data for the current tick */

	i = ch->playing_pos >> 8; 	/* i = playing_pos / 256 */
	/* haven't it finished playing the instrument? */
	if (i < ch->instr_len) {
		//n = ch->playing_pos >> 8;
		d = ch->loudness * ch->p_instr_data[i];
		d = d >> 8;
		blockbuf[bufidx] += d;
//		blockbuf[bufpos] += (ch->loudness * ch->p_instr_data[ch->playing_pos >> 8]) >> 8;

	//	ch->final_volume = (a * ch->loudness) >> 8;
	//	blockbuf[n] += ch->final_volume;
//		ch->y += ch->ticks_to_repeat;
//		ch->playing_pos = ch->y >> 8;
		ch->playing_pos += ch->freq_corrector;
	} 
	/* if we have finished and this instrument has to be repeated */
	else if (ch->instr_replen > 2) {
		i = ch->rep_playing_pos >> 8;
		d = ch->loudness * ch->p_instr_data[i];
		d = d >> 8;
		blockbuf[bufidx] += d;
//		ch->final_volume = (a * ch->loudness) >> 8;
		//blockbuf[n] += ch->final_volume;
		ch->rep_playing_pos += ch->freq_corrector;

		if (ch->rep_playing_pos >> 8 > ch->instr_reppos + ch->instr_replen - 1)
			ch->rep_playing_pos = ch->instr_reppos << 8;
	}
}

void *snd_play(void *arg) {
	int				ret;
#ifdef USE_THREAD
	while(!modctl.stop){
		sem_post(&sem_bin2);
		sem_wait(&sem_bin);
		ret = write(oss_driver.fd, buf2, BUFSIZE); 
	}
	pthread_exit(0);
#else
	ret = write(oss_driver.fd, buf2, BUFSIZE); 
#endif 
}

static int inst_range(int inst) {
	return(inst < 1 || inst > MAXINS ? 0 : inst);
}

void reset_note(struct channeltype *ch) {
	ch->freq_corrector = NTSC_AMIGA_MAGIC_NUM / (ch->final_period << 1);
	ch->freq_corrector = (ch->freq_corrector << 8) / MIXSPEED;
}

void reset_speed(void) {
	modctl.hertz = (modctl.bpm * 2) / 5;		/* Normally equals to 50 Hz */
	modctl.pal_vblank_speed = MIXSPEED / modctl.hertz;
}

void mp_setvolume(int vol) {
	modctl.volume = vol;
}

void refresh_channel_row(struct channeltype *ch, 
		struct modheadertype *mod_header, unsigned char modchannel[]) {

	int	period, instr;
	int	effect, effect_arg;
	
	extern char                         *instr_body_ptr[MAXINS];

	period = ((modchannel[0] & 0x0F) << 8) + modchannel[1];
	instr = (modchannel[0] & 0xF0) + (modchannel[2] >> 4);
	effect = modchannel[2] & 0x0F;
	effect_arg = modchannel[3];

	instr = inst_range(instr);

	ch->effect = FX_NOFX;			/* to ensure it is not a valid effect at first */
	ch->effect_argx = 0;
	ch->effect_argy = 0;

	if (instr) {
		ch->instr = instr;
		ch->loudness = mod_header->modinstr[ch->instr - 1].sampamp;
	}

	if (period && effect != 3 && effect != 5 &&
	(effect << 8) | (effect_arg & 0x0f0) != 0x0ed0) {
		ch->current_period = period;
		ch->final_period = period;
		reset_note(ch);

		ch->p_instr_data = instr_body_ptr[ch->instr - 1];
		ch->instr_len = mod_header->modinstr[ch->instr - 1].samplen;
		ch->playing_pos = 0;
		ch->instr_len = ch->instr_len <= 2 ? 0 : ch->instr_len;
		ch->instr_reppos = mod_header->modinstr[ch->instr - 1].sampreps;
		ch->instr_replen = mod_header->modinstr[ch->instr - 1].samprepl;
		/* if we have instrument repetion, then we must reconfigure 
		 * instrument size, i.e., when the instrument loop ends before
		 * the instrument end, we should *never* play the instrument all
		 * the way to end. */
		ch->instr_len = ch->instr_replen > 2 ? ch->instr_reppos + ch->instr_replen : ch->instr_len;
		ch->rep_playing_pos = ch->instr_reppos << 8;
	}		
	
	process_effect(ch, effect, effect_arg, period);
}

void refresh_row(int pos, int r) {
	int			pat;

	pat = mod_header.modarrangement[pos];
	refresh_channel_row(&channel1, &mod_header, &patterns[pat].row[r].channel1[0]);
	refresh_channel_row(&channel2, &mod_header, &patterns[pat].row[r].channel2[0]);
	refresh_channel_row(&channel3, &mod_header, &patterns[pat].row[r].channel3[0]);
	refresh_channel_row(&channel4, &mod_header, &patterns[pat].row[r].channel4[0]);
}


void mix_and_buf(void){
	static 		bufpos = 0;
	int			n = 0;

	if (oss_driver.channels > 1){        /* using a stereo card */
		channel_mix(&channel1, bufpos);
		channel_mix(&channel4, bufpos);
		bufpos++;
		channel_mix(&channel2, bufpos);
		channel_mix(&channel3, bufpos);
		bufpos++;
	}
//		else
//		{
//			channel_mix(&channel1, bufpos);
//			channel_mix(&channel2, bufpos);
//			channel_mix(&channel3, bufpos);
//			channel_mix(&channel4, bufpos);
//			bufpos++;
//		}
		
		/* buffer is full, lets dump it to dsp */
		if (bufpos >= BUFSIZE) {
			//sem_wait(&sem_bin2);
#ifdef USE_THREAD
			while (sem_trywait(&sem_bin2))
				usleep(100);
#endif
			for (n = 0; n < BUFSIZE; n++)
				buf2[n] = ((int)blockbuf[n] * modctl.volume / 100) - 128;
#ifdef USE_THREAD
			sem_post(&sem_bin);
#else
			snd_play(&blockbuf);
#endif 
			memset(blockbuf, 0, BUFSIZE * sizeof(int));   /* cleaning up buffer */
			bufpos = 0;
		}

}

void refresh_status_line(void) {
		
			/*n1 = channel1.final_volume;
			n2 = channel2.final_volume;
			n3 = channel3.final_volume;
			n4 = channel4.final_volume;
			n1 = n1 > 0 ? n1 : n1 * -1;
			n2 = n2 > 0 ? n2 : n2 * -1;
			n3 = n3 > 0 ? n3 : n3 * -1;
			n4 = n4 > 0 ? n4 : n4 * -1;
			n1 = n1 * 3 / 32;
			n2 = n2 * 3 / 32;
			n3 = n3 * 3 / 32;
			n4 = n4 * 3 / 32;

				printf("  %02d  %03d/%03d    %c   %c   %c   %c \r", line, arrpos, mod_header.modlen - 1, 
							blob[n1], blob[n2], blob[n3], blob[n4]);
				fflush(stdout);

			
	//			n1 = channel1.final_volume < n1 ? channel1.final_volume : n1;
	//			n2 = channel1.final_volume > n2 ? channel1.final_volume : n2;
	//			printf("n1 %d  n2 %d\n", n1, n2);
//			}
*/	
				printf("\r %02d    %03d   %02d  %03d/%03d", modctl.speed, modctl.bpm, modctl.row, modctl.pos, modctl.maxpos);
				fflush(stdout);
}


int mp_init(void) {
	int									result;
	int									i = 0;
	long int							tmp;
	int									ret = 0;

	modctl.bpm = DEFAULT_BPM;					/* Default Beats Per Minute */
	modctl.speed = DEFAULT_SPEED;				/* Num. of mod ticks per row */ 
	modctl.pos = 0;
	modctl.tick = 0;
	modctl.stop = 0;
	modctl.maxpos = mod_header.modlen - 1;
	modctl.row = 0;
	modctl.loop = 0;
	modctl.volume = 100;

	reset_speed();
	
	
	if ((blockbuf = (int *)malloc(BUFSIZE * sizeof(int))) == 0) {
		printf("Error allocating memory for instrument data.\n");
		ret = 1;
		goto mp_init_exit;
	}
	if ((buf2 = (char *)malloc(BUFSIZE)) == 0) {
		printf("Error allocating memory for instrument data.\n");
		ret = 1;
		goto mp_init_exit;
	}

	memset(blockbuf, 0, BUFSIZE * sizeof(int));   /* cleaning up buffer */
	
#ifdef USE_THREAD 
	tmp = sem_init(&sem_bin, 0, 0);
	tmp += sem_init(&sem_bin2, 0, 0);
    if (tmp != 0) {
        perror("Semaphore initialization failed");
        exit(EXIT_FAILURE);
    }
	tmp = pthread_attr_init(&mp_thread_attr);
//	tmp=pthread_attr_setinheritsched(&mp_thread_attr, PTHREAD_INHERIT_SCHED);
	tmp = pthread_attr_setschedpolicy(&mp_thread_attr, SCHED_OTHER);
	scheduling_value.sched_priority = sched_get_priority_min(SCHED_OTHER);
	tmp = pthread_attr_setschedparam(&mp_thread_attr, &scheduling_value);
	//tmp=pthread_create(&snd_thread, &mp_thread_attr, snd_play, snd_fd);

	tmp = pthread_create(&snd_thread, &mp_thread_attr, snd_play, NULL);

	if (tmp){
		printf("Thread creation error\n");
		ret = -1;
	}
#endif		

mp_init_exit:
	if (ret)
   		mp_stop();
	return (ret);
}

int mp_play(int ticks_to_play) {
	register int		n;
	register int		ticks_ctr = 0;
	static char 		n1, n2, n3, n4;
	static int			system_tick = 0;
	static int			lastrow = -1;
	static int			lastmodtick = -1;
	
	
	while (ticks_ctr < ticks_to_play) {
		if (modctl.row != lastrow) {
			refresh_row(modctl.pos, modctl.row);
	//		refresh_status_line();
			lastrow = modctl.row;
		}
		if (modctl.tick != lastmodtick) {
		/* lets do those effects that are tick based */
		if (modctl.tick) {				/* current tick > 1? */
			effect_tick(&channel1);
			effect_tick(&channel2);
			effect_tick(&channel3);
			effect_tick(&channel4);
		}
			lastmodtick = modctl.tick;
		}
		mix_and_buf();
		if (++system_tick > modctl.pal_vblank_speed) {
			system_tick = 0;
			modctl.tick++;
		}
		if (modctl.tick >= modctl.speed) {
			modctl.tick = 0;
			/* when doing effects that change current pattern (.i.e 0xB or 0xD),
			 * we don't incremet row here */
			if (modctl.leaving_block) {
				modctl.leaving_block = 0;
				lastrow = -1;   		/* forces row refresh */
			}
			else {
				modctl.row++;
			}
		}
		if (modctl.row > 63) {
			modctl.row = 0;
			modctl.pos++;
		}
		/* end of song? */
		if (modctl.pos > modctl.maxpos) {
			if (!modctl.loop) {  		/* should we do infinite loop? */	
				printf("\n"); 
				modctl.stop = 1; 
				break;
			}
			modctl.pos = 0;
		}
		ticks_ctr++;
	} /* end while */
	return (modctl.stop);
}

void mp_stop(void) {
	if (blockbuf)
		free(blockbuf);
	if (buf2)
		free(buf2);
	blockbuf = 0;
	buf2 = 0;
#ifdef USE_THREAD
	modctl.stop = 1;
	sem_post(&sem_bin);
	tmp = pthread_join(snd_thread, &thread_result);
	tmp = sem_destroy(&sem_bin);
#endif
}


